(function(mod) {
  if (typeof exports == "object" && typeof module == "object") // CommonJS
    mod(require("../../lib/codemirror"));
  else if (typeof define == "function" && define.amd) // AMD
    define(["../../lib/codemirror"], mod);
  else // Plain browser env
    mod(CodeMirror);
})(function(CodeMirror) {
"use strict";

    CodeMirror.defineMode("hsp", function(conf, parserConf) {
        var indentUnit = conf.indentUnit,
            statementIndentUnit = parserConf.statementIndentUnit || indentUnit,
            dontAlignCalls = parserConf.dontAlignCalls,
            keywords = parserConf.keywords || {},
            blockKeywords = parserConf.blockKeywords || {},
            atoms = parserConf.atoms || {},
            multiLineStrings = parserConf.multiLineStrings;

        // /hoge/ <- Regular Expression(Compiled)
        var operatorChar = /[+\-*&%=<>!|\/\\\^]/;
        var curPunc;
        function tokenBase(stream, state) {
            var ch = stream.next();
            if(ch == '"' || ch == "'") {
                state.tokenize = tokenString(ch);
                return state.tokenize(stream, state);
            }
            if(/[\[\]{}\(\),\.\:]/.test(ch)) {
                curPunc = ch;
                return null;
            }
            if(/\d/.test(ch)) {
                stream.eatWhile(/[\w\.]/);
                return "number";
            }
            if(ch == "/") {
                if(stream.eat("*")) {
                    state.tokenize = tokenComment;
                    return tokenComment(stream, state);
                }
                if(stream.eat("/")) {
                    stream.skipToEnd();
                    return "comment";
                }
            }
            if(ch == ";") {
                stream.eatWhile(/.*/);
                return "comment";
            }
            if(/\*/.test(ch)) {
                stream.eatWhile(/[A-Za-z0-9_]/);
                return "qualifier";
            }
            if(operatorChar.test(ch)) {
                stream.eatWhile(operatorChar);
                return "operator";
            }
            stream.eatWhile(/[\w\$_]/);
            var cur = stream.current();
            if(keywords.propertyIsEnumerable(cur)) {
                if(blockKeywords.propertyIsEnumerable(cur)) {
                    curPunc = "newstatement";
                }
                return "keyword";
            }
            if(atoms.propertyIsEnumerable(cur)) {
                return "atom";
            }
            return "variable";
        }

        function tokenString(quote) {
            return function(stream, state) {
                var escaped = false, next, end = false;
                while((next = stream.next()) != null) {
                    if(next == quote && !escaped) {
                        end = true;
                        break;
                    }
                    escaped = !escaped && next == "\\";
                }
                if(end || !(escaped || multiLineStrings)) {
                    state.tokenize = null;
                }
                return "string";
            };
        }

        function tokenComment(stream, state) {
            var maybeEnd = false, ch;
            while(ch = stream.next()) {
                if(ch == "/" && maybeEnd) {
                    state.tokenize = null;
                    break;
                }
                maybeEnd = (ch == "*");
            }
            return "comment";
        }

        function Context(indented, column, type, align, prev) {
            this.indented = indented;
            this.column = column;
            this.type = type;
            this.align = align;
            this.prev = prev;
        }

        function pushContext(state, column, type) {
            var indent = state.indented;
            if(state.context && state.context.type == "statement") {
                indent = state.context.indented;
            }
            return state.context = new Context(indent, column, type, null, state.context);
        }

        function popContext(state) {
            var t = state.context.type;
            if(t == ")" || t == "]" || t == "}") {
                state.indented = state.context.indented;
            }
            return state.context = state.context.prev;
        }

        return {
            startState: function(basecolumn) {
                return {
                    tokenize: null,
                    context: new Context((basecolumn || 0) - indentUnit, 0, "top", false),
                    indented: 0,
                    startOfLine: true
                };
            },
            token: function(stream, state) {
                var ctx = state.context;
                if(stream.sol()) {
                    if(ctx.align == null) {
                        ctx.align= false;
                    }
                    state.indented = stream.indentation();
                    state.startOfLine = true;
                }
                if(stream.eatSpace()) {
                    return null;
                }
                curPunc = null;
                var style = (state.tokenize || tokenBase)(stream, state);
                if(style == "comment" || style == "meta") {
                    return style;
                }
                if(ctx.align == null) {
                    ctx.align = true;
                }
                if((curPunc == ";" || curPunc == ":" || curPunc == ",") && ctx.type == "statement") {
                    popContext(state);
                } else if(curPunc == "{") {
                    pushContext(state, stream.column(), "}");
                } else if(curPunc == "[") {
                    pushContext(state, stream.column(), "]");
                } else if(curPunc == "(") {
                    pushContext(state, stream.column(), ")");
                } else if(curPunc == "}") {
                    while(ctx.type == "statement") {
                        ctx = popContext(state);
                    }
                    if(ctx.type == "}") {
                        ctx = popContext(state);
                    }
                    while(ctx.type == "statement") {
                        ctx = popContext(state);
                    }
                } else if(curPunc == ctx.type) {
                    popContext(state);
                } else if(((ctx.type == "}" || ctx.type == "top") && curPunc != ';') || (ctx.type == "statement" && curPunc == "newstatement")) {
                    pushContext(state, stream.column(), "statement");
                }
                state.startOfLine = false;
                return style;
            },
            indent: function(state, textAfter) {
                if(state.tokenize != tokenBase && state.tokenize != null) {
                    return CodeMirror.Pass;
                }
                var ctx = state.context, firstChar = textAfter && textAfter.charAt(0);
                if(ctx.type == "statement" && firstChar == "}") {
                    ctx = ctx.prev;
                }
                var closing = firstChar == ctx.type;
                if(ctx.type == "statement") {
                    return ctx.indented + (firstChar == "{" ? 0 : statementIndentUnit);
                } else if(ctx.align && (!dontAlignCalls || ctx.type != ")")) {
                    return ctx.column + (closing ? 0 : 1);
                } else if(ctx.type == ")" && !closing) {
                    return ctx.indented + statementIndentUnit;
                } else {
                    return ctx.indented + (closing ? 0 : indentUnit);
                }
            },
            electricChars: "{}",
            blockCommentStart: "/*",
            blockCommentEnd: "*/",
            lineComment: "//",
            fold: "brace"
        };
        
    });

    function words(str) {
        var obj = {}, words = str.split(" ");
        for(var i = 0; i < words.length; ++i) obj[words[i]] = true;
        return obj;
    }

    var hspKeywords = "_break _contine abs absf alloc assert atan await axobj bcopy bgscr bload bmpsabe boxf break" +
                   " bsave buffer button case celdiv celload celput chdir chdpm chgdisp chkbox circle " +
                   " clrobj cls cnvwtos cnvstow color combox comevarg comevdisp comevent comres continue cos ddim" +
                   " default delete delcom delmod dialog dim dirinfo dirlist do double dup dupptr else end" + 
                   " exec exgoto exist expf font for foreach gcopy getstr gettime getkey getpath ginfo gmode gosub" +
                   " goto gradf grect groll grotate gsel gsquare gzoom hsvcolor if input instr int ldim length" + 
                   " length2 length3 length4 libptr limit limitf line listbox logf logmes loop lpeek lpoke mcall" + 
                   " mci mes mesbox memcpy memexpand memfile memset mkdir mmload mmplay mmstop mouse mref" +
                   " newcom newlab newmod next noteadd notedel noteget noteinfo noteload notesave notesel" +
                   " noteunsel objenable objimage objinfo objmode objprm objsel objsize objskip on oncick oncmd" +
                   " onerror onexit onkey palcolor palette peek pget picload poke pos powf print pset qureycom" + 
                   " randomize redraw repeat resume return rnd run sarrayconv screen sdim sin split sqrt stick" +
                   " stop str strf strlen strmid strtrim strrep swbreak swend switch syscolor sysfont sysinfo" +
                   " title until varptr vartype varuse wait wend while width winobj wpeek wpoke yield" +
                   " #addition #aht #ahtmes #cfunc #cmd #cmpopt #comfunc #const #defcfunc #deffunc #define #else" +
                   " #endif #enum #epack #func #global #if #ifdef #ifndef #include #modcfunc #modfunc #modinit" +
                   " #modterm #module #pack #packopt #regcmd #runtime #undef #usecom #uselib";
    var hspBlockKeywords = "case default else for foreach if repeat switch while";
    var hspAtoms = "true false null";

    var hspConfig = {
        name: "hsp",
        keywords: words(hspKeywords),
        blockKeywords: words(hspBlockKeywords),
        atoms: words(hspAtoms),
        helperType: "text/x-hsp",
        modeProps: {fold: ["brace", "include"]}
    };

    var words_ = [];

    //register
    function add(obj) {
        if (obj) for (var prop in obj) if (obj.hasOwnProperty(prop)) {
            words_.push(prop);
        }
    }

    add(hspConfig.keywords);
    add(hspConfig.atoms);

    CodeMirror.registerHelper("hintWords", "text/x-hsp", words_);
    CodeMirror.defineMIME("text/x-hsp", hspConfig);

});